
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "app-enumtypes.h"
#include "gstappsrc.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gstappsrc.h" */

GType
gst_app_stream_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_APP_STREAM_TYPE_STREAM), "GST_APP_STREAM_TYPE_STREAM", "stream" },
    { C_ENUM(GST_APP_STREAM_TYPE_SEEKABLE), "GST_APP_STREAM_TYPE_SEEKABLE", "seekable" },
    { C_ENUM(GST_APP_STREAM_TYPE_RANDOM_ACCESS), "GST_APP_STREAM_TYPE_RANDOM_ACCESS", "random-access" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstAppStreamType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_app_leaky_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_APP_LEAKY_TYPE_NONE), "GST_APP_LEAKY_TYPE_NONE", "none" },
    { C_ENUM(GST_APP_LEAKY_TYPE_UPSTREAM), "GST_APP_LEAKY_TYPE_UPSTREAM", "upstream" },
    { C_ENUM(GST_APP_LEAKY_TYPE_DOWNSTREAM), "GST_APP_LEAKY_TYPE_DOWNSTREAM", "downstream" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstAppLeakyType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* Generated data ends here */

