
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "pbutils-enumtypes.h"
#include "pbutils.h"
#include "pbutils-prelude.h"
#include "codec-utils.h"
#include "descriptions.h"
#include "encoding-profile.h"
#include "encoding-target.h"
#include "install-plugins.h"
#include "missing-plugins.h"
#include "gstdiscoverer.h"
#include "gstaudiovisualizer.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "descriptions.h" */

GType
gst_pb_utils_caps_description_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_PBUTILS_CAPS_DESCRIPTION_FLAG_CONTAINER), "GST_PBUTILS_CAPS_DESCRIPTION_FLAG_CONTAINER", "container" },
    { C_FLAGS(GST_PBUTILS_CAPS_DESCRIPTION_FLAG_AUDIO), "GST_PBUTILS_CAPS_DESCRIPTION_FLAG_AUDIO", "audio" },
    { C_FLAGS(GST_PBUTILS_CAPS_DESCRIPTION_FLAG_VIDEO), "GST_PBUTILS_CAPS_DESCRIPTION_FLAG_VIDEO", "video" },
    { C_FLAGS(GST_PBUTILS_CAPS_DESCRIPTION_FLAG_IMAGE), "GST_PBUTILS_CAPS_DESCRIPTION_FLAG_IMAGE", "image" },
    { C_FLAGS(GST_PBUTILS_CAPS_DESCRIPTION_FLAG_SUBTITLE), "GST_PBUTILS_CAPS_DESCRIPTION_FLAG_SUBTITLE", "subtitle" },
    { C_FLAGS(GST_PBUTILS_CAPS_DESCRIPTION_FLAG_TAG), "GST_PBUTILS_CAPS_DESCRIPTION_FLAG_TAG", "tag" },
    { C_FLAGS(GST_PBUTILS_CAPS_DESCRIPTION_FLAG_GENERIC), "GST_PBUTILS_CAPS_DESCRIPTION_FLAG_GENERIC", "generic" },
    { C_FLAGS(GST_PBUTILS_CAPS_DESCRIPTION_FLAG_METADATA), "GST_PBUTILS_CAPS_DESCRIPTION_FLAG_METADATA", "metadata" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstPbUtilsCapsDescriptionFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gstaudiovisualizer.h" */

GType
gst_audio_visualizer_shader_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_NONE), "GST_AUDIO_VISUALIZER_SHADER_NONE", "none" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE), "GST_AUDIO_VISUALIZER_SHADER_FADE", "fade" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_UP), "GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_UP", "fade-and-move-up" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_DOWN), "GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_DOWN", "fade-and-move-down" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_LEFT), "GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_LEFT", "fade-and-move-left" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_RIGHT), "GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_RIGHT", "fade-and-move-right" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_HORIZ_OUT), "GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_HORIZ_OUT", "fade-and-move-horiz-out" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_HORIZ_IN), "GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_HORIZ_IN", "fade-and-move-horiz-in" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_VERT_OUT), "GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_VERT_OUT", "fade-and-move-vert-out" },
    { C_ENUM(GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_VERT_IN), "GST_AUDIO_VISUALIZER_SHADER_FADE_AND_MOVE_VERT_IN", "fade-and-move-vert-in" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstAudioVisualizerShader"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gstdiscoverer.h" */

GType
gst_discoverer_result_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_DISCOVERER_OK), "GST_DISCOVERER_OK", "ok" },
    { C_ENUM(GST_DISCOVERER_URI_INVALID), "GST_DISCOVERER_URI_INVALID", "uri-invalid" },
    { C_ENUM(GST_DISCOVERER_ERROR), "GST_DISCOVERER_ERROR", "error" },
    { C_ENUM(GST_DISCOVERER_TIMEOUT), "GST_DISCOVERER_TIMEOUT", "timeout" },
    { C_ENUM(GST_DISCOVERER_BUSY), "GST_DISCOVERER_BUSY", "busy" },
    { C_ENUM(GST_DISCOVERER_MISSING_PLUGINS), "GST_DISCOVERER_MISSING_PLUGINS", "missing-plugins" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstDiscovererResult"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_discoverer_serialize_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_DISCOVERER_SERIALIZE_BASIC), "GST_DISCOVERER_SERIALIZE_BASIC", "basic" },
    { C_FLAGS(GST_DISCOVERER_SERIALIZE_CAPS), "GST_DISCOVERER_SERIALIZE_CAPS", "caps" },
    { C_FLAGS(GST_DISCOVERER_SERIALIZE_TAGS), "GST_DISCOVERER_SERIALIZE_TAGS", "tags" },
    { C_FLAGS(GST_DISCOVERER_SERIALIZE_MISC), "GST_DISCOVERER_SERIALIZE_MISC", "misc" },
    { C_FLAGS(GST_DISCOVERER_SERIALIZE_ALL), "GST_DISCOVERER_SERIALIZE_ALL", "all" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstDiscovererSerializeFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "install-plugins.h" */

GType
gst_install_plugins_return_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_INSTALL_PLUGINS_SUCCESS), "GST_INSTALL_PLUGINS_SUCCESS", "success" },
    { C_ENUM(GST_INSTALL_PLUGINS_NOT_FOUND), "GST_INSTALL_PLUGINS_NOT_FOUND", "not-found" },
    { C_ENUM(GST_INSTALL_PLUGINS_ERROR), "GST_INSTALL_PLUGINS_ERROR", "error" },
    { C_ENUM(GST_INSTALL_PLUGINS_PARTIAL_SUCCESS), "GST_INSTALL_PLUGINS_PARTIAL_SUCCESS", "partial-success" },
    { C_ENUM(GST_INSTALL_PLUGINS_USER_ABORT), "GST_INSTALL_PLUGINS_USER_ABORT", "user-abort" },
    { C_ENUM(GST_INSTALL_PLUGINS_CRASHED), "GST_INSTALL_PLUGINS_CRASHED", "crashed" },
    { C_ENUM(GST_INSTALL_PLUGINS_INVALID), "GST_INSTALL_PLUGINS_INVALID", "invalid" },
    { C_ENUM(GST_INSTALL_PLUGINS_STARTED_OK), "GST_INSTALL_PLUGINS_STARTED_OK", "started-ok" },
    { C_ENUM(GST_INSTALL_PLUGINS_INTERNAL_FAILURE), "GST_INSTALL_PLUGINS_INTERNAL_FAILURE", "internal-failure" },
    { C_ENUM(GST_INSTALL_PLUGINS_HELPER_MISSING), "GST_INSTALL_PLUGINS_HELPER_MISSING", "helper-missing" },
    { C_ENUM(GST_INSTALL_PLUGINS_INSTALL_IN_PROGRESS), "GST_INSTALL_PLUGINS_INSTALL_IN_PROGRESS", "install-in-progress" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstInstallPluginsReturn"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* Generated data ends here */

