
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "video-enumtypes.h"
#include "video.h"
#include "video-anc.h"
#include "video-format.h"
#include "video-frame.h"
#include "video-chroma.h"
#include "video-color.h"
#include "video-converter.h"
#include "video-dither.h"
#include "video-info.h"
#include "video-overlay-composition.h"
#include "video-resampler.h"
#include "video-scaler.h"
#include "video-tile.h"
#include "gstvideometa.h"
#include "gstvideotimecode.h"
#include "gstvideoutils.h"
#include "gstvideoencoder.h"
#include "gstvideodecoder.h"
#include "colorbalance.h"
#include "navigation.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "colorbalance.h" */

GType
gst_color_balance_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_COLOR_BALANCE_HARDWARE), "GST_COLOR_BALANCE_HARDWARE", "hardware" },
    { C_ENUM(GST_COLOR_BALANCE_SOFTWARE), "GST_COLOR_BALANCE_SOFTWARE", "software" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstColorBalanceType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gstvideodecoder.h" */

GType
gst_video_decoder_request_sync_point_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_DECODER_REQUEST_SYNC_POINT_DISCARD_INPUT), "GST_VIDEO_DECODER_REQUEST_SYNC_POINT_DISCARD_INPUT", "discard-input" },
    { C_FLAGS(GST_VIDEO_DECODER_REQUEST_SYNC_POINT_CORRUPT_OUTPUT), "GST_VIDEO_DECODER_REQUEST_SYNC_POINT_CORRUPT_OUTPUT", "corrupt-output" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoDecoderRequestSyncPointFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gstvideometa.h" */

GType
gst_video_gl_texture_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_GL_TEXTURE_TYPE_LUMINANCE), "GST_VIDEO_GL_TEXTURE_TYPE_LUMINANCE", "luminance" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_TYPE_LUMINANCE_ALPHA), "GST_VIDEO_GL_TEXTURE_TYPE_LUMINANCE_ALPHA", "luminance-alpha" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_TYPE_RGB16), "GST_VIDEO_GL_TEXTURE_TYPE_RGB16", "rgb16" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_TYPE_RGB), "GST_VIDEO_GL_TEXTURE_TYPE_RGB", "rgb" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_TYPE_RGBA), "GST_VIDEO_GL_TEXTURE_TYPE_RGBA", "rgba" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_TYPE_R), "GST_VIDEO_GL_TEXTURE_TYPE_R", "r" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_TYPE_RG), "GST_VIDEO_GL_TEXTURE_TYPE_RG", "rg" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoGLTextureType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_gl_texture_orientation_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_GL_TEXTURE_ORIENTATION_X_NORMAL_Y_NORMAL), "GST_VIDEO_GL_TEXTURE_ORIENTATION_X_NORMAL_Y_NORMAL", "normal-y-normal" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_ORIENTATION_X_NORMAL_Y_FLIP), "GST_VIDEO_GL_TEXTURE_ORIENTATION_X_NORMAL_Y_FLIP", "normal-y-flip" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_ORIENTATION_X_FLIP_Y_NORMAL), "GST_VIDEO_GL_TEXTURE_ORIENTATION_X_FLIP_Y_NORMAL", "flip-y-normal" },
    { C_ENUM(GST_VIDEO_GL_TEXTURE_ORIENTATION_X_FLIP_Y_FLIP), "GST_VIDEO_GL_TEXTURE_ORIENTATION_X_FLIP_Y_FLIP", "flip-y-flip" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoGLTextureOrientation"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gstvideotimecode.h" */

GType
gst_video_time_code_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_TIME_CODE_FLAGS_NONE), "GST_VIDEO_TIME_CODE_FLAGS_NONE", "none" },
    { C_FLAGS(GST_VIDEO_TIME_CODE_FLAGS_DROP_FRAME), "GST_VIDEO_TIME_CODE_FLAGS_DROP_FRAME", "drop-frame" },
    { C_FLAGS(GST_VIDEO_TIME_CODE_FLAGS_INTERLACED), "GST_VIDEO_TIME_CODE_FLAGS_INTERLACED", "interlaced" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoTimeCodeFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "gstvideoutils.h" */

GType
gst_video_codec_frame_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_CODEC_FRAME_FLAG_DECODE_ONLY), "GST_VIDEO_CODEC_FRAME_FLAG_DECODE_ONLY", "decode-only" },
    { C_FLAGS(GST_VIDEO_CODEC_FRAME_FLAG_SYNC_POINT), "GST_VIDEO_CODEC_FRAME_FLAG_SYNC_POINT", "sync-point" },
    { C_FLAGS(GST_VIDEO_CODEC_FRAME_FLAG_FORCE_KEYFRAME), "GST_VIDEO_CODEC_FRAME_FLAG_FORCE_KEYFRAME", "force-keyframe" },
    { C_FLAGS(GST_VIDEO_CODEC_FRAME_FLAG_FORCE_KEYFRAME_HEADERS), "GST_VIDEO_CODEC_FRAME_FLAG_FORCE_KEYFRAME_HEADERS", "force-keyframe-headers" },
    { C_FLAGS(GST_VIDEO_CODEC_FRAME_FLAG_XLNX_ALREADY_PUSHED), "GST_VIDEO_CODEC_FRAME_FLAG_XLNX_ALREADY_PUSHED", "xlnx-already-pushed" },
    { C_FLAGS(GST_VIDEO_CODEC_FRAME_FLAG_CORRUPTED), "GST_VIDEO_CODEC_FRAME_FLAG_CORRUPTED", "corrupted" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoCodecFrameFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "navigation.h" */

GType
gst_navigation_modifier_type_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_NAVIGATION_MODIFIER_NONE), "GST_NAVIGATION_MODIFIER_NONE", "none" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_SHIFT_MASK), "GST_NAVIGATION_MODIFIER_SHIFT_MASK", "shift-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_LOCK_MASK), "GST_NAVIGATION_MODIFIER_LOCK_MASK", "lock-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_CONTROL_MASK), "GST_NAVIGATION_MODIFIER_CONTROL_MASK", "control-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_MOD1_MASK), "GST_NAVIGATION_MODIFIER_MOD1_MASK", "mod1-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_MOD2_MASK), "GST_NAVIGATION_MODIFIER_MOD2_MASK", "mod2-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_MOD3_MASK), "GST_NAVIGATION_MODIFIER_MOD3_MASK", "mod3-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_MOD4_MASK), "GST_NAVIGATION_MODIFIER_MOD4_MASK", "mod4-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_MOD5_MASK), "GST_NAVIGATION_MODIFIER_MOD5_MASK", "mod5-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_BUTTON1_MASK), "GST_NAVIGATION_MODIFIER_BUTTON1_MASK", "button1-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_BUTTON2_MASK), "GST_NAVIGATION_MODIFIER_BUTTON2_MASK", "button2-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_BUTTON3_MASK), "GST_NAVIGATION_MODIFIER_BUTTON3_MASK", "button3-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_BUTTON4_MASK), "GST_NAVIGATION_MODIFIER_BUTTON4_MASK", "button4-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_BUTTON5_MASK), "GST_NAVIGATION_MODIFIER_BUTTON5_MASK", "button5-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_SUPER_MASK), "GST_NAVIGATION_MODIFIER_SUPER_MASK", "super-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_HYPER_MASK), "GST_NAVIGATION_MODIFIER_HYPER_MASK", "hyper-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_META_MASK), "GST_NAVIGATION_MODIFIER_META_MASK", "meta-mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_MASK), "GST_NAVIGATION_MODIFIER_MASK", "mask" },
    { C_FLAGS(GST_NAVIGATION_MODIFIER_META_MASK), "GST_NAVIGATION_MODIFIER_META_MASK", "meta-mask" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstNavigationModifierType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_navigation_command_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_NAVIGATION_COMMAND_INVALID), "GST_NAVIGATION_COMMAND_INVALID", "invalid" },
    { C_ENUM(GST_NAVIGATION_COMMAND_MENU1), "GST_NAVIGATION_COMMAND_MENU1", "menu1" },
    { C_ENUM(GST_NAVIGATION_COMMAND_MENU2), "GST_NAVIGATION_COMMAND_MENU2", "menu2" },
    { C_ENUM(GST_NAVIGATION_COMMAND_MENU3), "GST_NAVIGATION_COMMAND_MENU3", "menu3" },
    { C_ENUM(GST_NAVIGATION_COMMAND_MENU4), "GST_NAVIGATION_COMMAND_MENU4", "menu4" },
    { C_ENUM(GST_NAVIGATION_COMMAND_MENU5), "GST_NAVIGATION_COMMAND_MENU5", "menu5" },
    { C_ENUM(GST_NAVIGATION_COMMAND_MENU6), "GST_NAVIGATION_COMMAND_MENU6", "menu6" },
    { C_ENUM(GST_NAVIGATION_COMMAND_MENU7), "GST_NAVIGATION_COMMAND_MENU7", "menu7" },
    { C_ENUM(GST_NAVIGATION_COMMAND_LEFT), "GST_NAVIGATION_COMMAND_LEFT", "left" },
    { C_ENUM(GST_NAVIGATION_COMMAND_RIGHT), "GST_NAVIGATION_COMMAND_RIGHT", "right" },
    { C_ENUM(GST_NAVIGATION_COMMAND_UP), "GST_NAVIGATION_COMMAND_UP", "up" },
    { C_ENUM(GST_NAVIGATION_COMMAND_DOWN), "GST_NAVIGATION_COMMAND_DOWN", "down" },
    { C_ENUM(GST_NAVIGATION_COMMAND_ACTIVATE), "GST_NAVIGATION_COMMAND_ACTIVATE", "activate" },
    { C_ENUM(GST_NAVIGATION_COMMAND_PREV_ANGLE), "GST_NAVIGATION_COMMAND_PREV_ANGLE", "prev-angle" },
    { C_ENUM(GST_NAVIGATION_COMMAND_NEXT_ANGLE), "GST_NAVIGATION_COMMAND_NEXT_ANGLE", "next-angle" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstNavigationCommand"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_navigation_query_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_NAVIGATION_QUERY_INVALID), "GST_NAVIGATION_QUERY_INVALID", "invalid" },
    { C_ENUM(GST_NAVIGATION_QUERY_COMMANDS), "GST_NAVIGATION_QUERY_COMMANDS", "commands" },
    { C_ENUM(GST_NAVIGATION_QUERY_ANGLES), "GST_NAVIGATION_QUERY_ANGLES", "angles" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstNavigationQueryType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_navigation_message_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_NAVIGATION_MESSAGE_INVALID), "GST_NAVIGATION_MESSAGE_INVALID", "invalid" },
    { C_ENUM(GST_NAVIGATION_MESSAGE_MOUSE_OVER), "GST_NAVIGATION_MESSAGE_MOUSE_OVER", "mouse-over" },
    { C_ENUM(GST_NAVIGATION_MESSAGE_COMMANDS_CHANGED), "GST_NAVIGATION_MESSAGE_COMMANDS_CHANGED", "commands-changed" },
    { C_ENUM(GST_NAVIGATION_MESSAGE_ANGLES_CHANGED), "GST_NAVIGATION_MESSAGE_ANGLES_CHANGED", "angles-changed" },
    { C_ENUM(GST_NAVIGATION_MESSAGE_EVENT), "GST_NAVIGATION_MESSAGE_EVENT", "event" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstNavigationMessageType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_navigation_event_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_NAVIGATION_EVENT_INVALID), "GST_NAVIGATION_EVENT_INVALID", "invalid" },
    { C_ENUM(GST_NAVIGATION_EVENT_KEY_PRESS), "GST_NAVIGATION_EVENT_KEY_PRESS", "key-press" },
    { C_ENUM(GST_NAVIGATION_EVENT_KEY_RELEASE), "GST_NAVIGATION_EVENT_KEY_RELEASE", "key-release" },
    { C_ENUM(GST_NAVIGATION_EVENT_MOUSE_BUTTON_PRESS), "GST_NAVIGATION_EVENT_MOUSE_BUTTON_PRESS", "mouse-button-press" },
    { C_ENUM(GST_NAVIGATION_EVENT_MOUSE_BUTTON_RELEASE), "GST_NAVIGATION_EVENT_MOUSE_BUTTON_RELEASE", "mouse-button-release" },
    { C_ENUM(GST_NAVIGATION_EVENT_MOUSE_MOVE), "GST_NAVIGATION_EVENT_MOUSE_MOVE", "mouse-move" },
    { C_ENUM(GST_NAVIGATION_EVENT_COMMAND), "GST_NAVIGATION_EVENT_COMMAND", "command" },
    { C_ENUM(GST_NAVIGATION_EVENT_MOUSE_SCROLL), "GST_NAVIGATION_EVENT_MOUSE_SCROLL", "mouse-scroll" },
    { C_ENUM(GST_NAVIGATION_EVENT_TOUCH_DOWN), "GST_NAVIGATION_EVENT_TOUCH_DOWN", "touch-down" },
    { C_ENUM(GST_NAVIGATION_EVENT_TOUCH_MOTION), "GST_NAVIGATION_EVENT_TOUCH_MOTION", "touch-motion" },
    { C_ENUM(GST_NAVIGATION_EVENT_TOUCH_UP), "GST_NAVIGATION_EVENT_TOUCH_UP", "touch-up" },
    { C_ENUM(GST_NAVIGATION_EVENT_TOUCH_FRAME), "GST_NAVIGATION_EVENT_TOUCH_FRAME", "touch-frame" },
    { C_ENUM(GST_NAVIGATION_EVENT_TOUCH_CANCEL), "GST_NAVIGATION_EVENT_TOUCH_CANCEL", "touch-cancel" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstNavigationEventType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-anc.h" */

GType
gst_video_ancillary_did_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_UNDEFINED), "GST_VIDEO_ANCILLARY_DID_UNDEFINED", "undefined" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_DELETION), "GST_VIDEO_ANCILLARY_DID_DELETION", "deletion" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_3G_AUDIO_DATA_FIRST), "GST_VIDEO_ANCILLARY_DID_HANC_3G_AUDIO_DATA_FIRST", "hanc-3g-audio-data-first" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_3G_AUDIO_DATA_LAST), "GST_VIDEO_ANCILLARY_DID_HANC_3G_AUDIO_DATA_LAST", "hanc-3g-audio-data-last" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_HDTV_AUDIO_DATA_FIRST), "GST_VIDEO_ANCILLARY_DID_HANC_HDTV_AUDIO_DATA_FIRST", "hanc-hdtv-audio-data-first" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_HDTV_AUDIO_DATA_LAST), "GST_VIDEO_ANCILLARY_DID_HANC_HDTV_AUDIO_DATA_LAST", "hanc-hdtv-audio-data-last" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_SDTV_AUDIO_DATA_1_FIRST), "GST_VIDEO_ANCILLARY_DID_HANC_SDTV_AUDIO_DATA_1_FIRST", "hanc-sdtv-audio-data-1-first" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_SDTV_AUDIO_DATA_1_LAST), "GST_VIDEO_ANCILLARY_DID_HANC_SDTV_AUDIO_DATA_1_LAST", "hanc-sdtv-audio-data-1-last" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_CAMERA_POSITION), "GST_VIDEO_ANCILLARY_DID_CAMERA_POSITION", "camera-position" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_ERROR_DETECTION), "GST_VIDEO_ANCILLARY_DID_HANC_ERROR_DETECTION", "hanc-error-detection" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_SDTV_AUDIO_DATA_2_FIRST), "GST_VIDEO_ANCILLARY_DID_HANC_SDTV_AUDIO_DATA_2_FIRST", "hanc-sdtv-audio-data-2-first" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID_HANC_SDTV_AUDIO_DATA_2_LAST), "GST_VIDEO_ANCILLARY_DID_HANC_SDTV_AUDIO_DATA_2_LAST", "hanc-sdtv-audio-data-2-last" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoAncillaryDID"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_ancillary_di_d16_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_ANCILLARY_DID16_S334_EIA_708), "GST_VIDEO_ANCILLARY_DID16_S334_EIA_708", "s334-eia-708" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID16_S334_EIA_608), "GST_VIDEO_ANCILLARY_DID16_S334_EIA_608", "s334-eia-608" },
    { C_ENUM(GST_VIDEO_ANCILLARY_DID16_S2016_3_AFD_BAR), "GST_VIDEO_ANCILLARY_DID16_S2016_3_AFD_BAR", "s2016-3-afd-bar" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoAncillaryDID16"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_afd_value_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_AFD_UNAVAILABLE), "GST_VIDEO_AFD_UNAVAILABLE", "unavailable" },
    { C_ENUM(GST_VIDEO_AFD_16_9_TOP_ALIGNED), "GST_VIDEO_AFD_16_9_TOP_ALIGNED", "16-9-top-aligned" },
    { C_ENUM(GST_VIDEO_AFD_14_9_TOP_ALIGNED), "GST_VIDEO_AFD_14_9_TOP_ALIGNED", "14-9-top-aligned" },
    { C_ENUM(GST_VIDEO_AFD_GREATER_THAN_16_9), "GST_VIDEO_AFD_GREATER_THAN_16_9", "greater-than-16-9" },
    { C_ENUM(GST_VIDEO_AFD_4_3_FULL_16_9_FULL), "GST_VIDEO_AFD_4_3_FULL_16_9_FULL", "4-3-full-16-9-full" },
    { C_ENUM(GST_VIDEO_AFD_4_3_FULL_4_3_PILLAR), "GST_VIDEO_AFD_4_3_FULL_4_3_PILLAR", "4-3-full-4-3-pillar" },
    { C_ENUM(GST_VIDEO_AFD_16_9_LETTER_16_9_FULL), "GST_VIDEO_AFD_16_9_LETTER_16_9_FULL", "16-9-letter-16-9-full" },
    { C_ENUM(GST_VIDEO_AFD_14_9_LETTER_14_9_PILLAR), "GST_VIDEO_AFD_14_9_LETTER_14_9_PILLAR", "14-9-letter-14-9-pillar" },
    { C_ENUM(GST_VIDEO_AFD_4_3_FULL_14_9_CENTER), "GST_VIDEO_AFD_4_3_FULL_14_9_CENTER", "4-3-full-14-9-center" },
    { C_ENUM(GST_VIDEO_AFD_16_9_LETTER_14_9_CENTER), "GST_VIDEO_AFD_16_9_LETTER_14_9_CENTER", "16-9-letter-14-9-center" },
    { C_ENUM(GST_VIDEO_AFD_16_9_LETTER_4_3_CENTER), "GST_VIDEO_AFD_16_9_LETTER_4_3_CENTER", "16-9-letter-4-3-center" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoAFDValue"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_afd_spec_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_AFD_SPEC_DVB_ETSI), "GST_VIDEO_AFD_SPEC_DVB_ETSI", "dvb-etsi" },
    { C_ENUM(GST_VIDEO_AFD_SPEC_ATSC_A53), "GST_VIDEO_AFD_SPEC_ATSC_A53", "atsc-a53" },
    { C_ENUM(GST_VIDEO_AFD_SPEC_SMPTE_ST2016_1), "GST_VIDEO_AFD_SPEC_SMPTE_ST2016_1", "smpte-st2016-1" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoAFDSpec"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_caption_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_CAPTION_TYPE_UNKNOWN), "GST_VIDEO_CAPTION_TYPE_UNKNOWN", "unknown" },
    { C_ENUM(GST_VIDEO_CAPTION_TYPE_CEA608_RAW), "GST_VIDEO_CAPTION_TYPE_CEA608_RAW", "cea608-raw" },
    { C_ENUM(GST_VIDEO_CAPTION_TYPE_CEA608_S334_1A), "GST_VIDEO_CAPTION_TYPE_CEA608_S334_1A", "cea608-s334-1a" },
    { C_ENUM(GST_VIDEO_CAPTION_TYPE_CEA708_RAW), "GST_VIDEO_CAPTION_TYPE_CEA708_RAW", "cea708-raw" },
    { C_ENUM(GST_VIDEO_CAPTION_TYPE_CEA708_CDP), "GST_VIDEO_CAPTION_TYPE_CEA708_CDP", "cea708-cdp" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoCaptionType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_vbi_parser_result_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_VBI_PARSER_RESULT_DONE), "GST_VIDEO_VBI_PARSER_RESULT_DONE", "done" },
    { C_ENUM(GST_VIDEO_VBI_PARSER_RESULT_OK), "GST_VIDEO_VBI_PARSER_RESULT_OK", "ok" },
    { C_ENUM(GST_VIDEO_VBI_PARSER_RESULT_ERROR), "GST_VIDEO_VBI_PARSER_RESULT_ERROR", "error" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoVBIParserResult"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-chroma.h" */

GType
gst_video_chroma_site_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_UNKNOWN), "GST_VIDEO_CHROMA_SITE_UNKNOWN", "unknown" },
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_NONE), "GST_VIDEO_CHROMA_SITE_NONE", "none" },
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_H_COSITED), "GST_VIDEO_CHROMA_SITE_H_COSITED", "h-cosited" },
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_V_COSITED), "GST_VIDEO_CHROMA_SITE_V_COSITED", "v-cosited" },
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_ALT_LINE), "GST_VIDEO_CHROMA_SITE_ALT_LINE", "alt-line" },
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_COSITED), "GST_VIDEO_CHROMA_SITE_COSITED", "cosited" },
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_JPEG), "GST_VIDEO_CHROMA_SITE_JPEG", "jpeg" },
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_MPEG2), "GST_VIDEO_CHROMA_SITE_MPEG2", "mpeg2" },
    { C_FLAGS(GST_VIDEO_CHROMA_SITE_DV), "GST_VIDEO_CHROMA_SITE_DV", "dv" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoChromaSite"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_chroma_method_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_CHROMA_METHOD_NEAREST), "GST_VIDEO_CHROMA_METHOD_NEAREST", "nearest" },
    { C_ENUM(GST_VIDEO_CHROMA_METHOD_LINEAR), "GST_VIDEO_CHROMA_METHOD_LINEAR", "linear" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoChromaMethod"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_chroma_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_CHROMA_FLAG_NONE), "GST_VIDEO_CHROMA_FLAG_NONE", "none" },
    { C_FLAGS(GST_VIDEO_CHROMA_FLAG_INTERLACED), "GST_VIDEO_CHROMA_FLAG_INTERLACED", "interlaced" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoChromaFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-color.h" */

GType
gst_video_color_range_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_COLOR_RANGE_UNKNOWN), "GST_VIDEO_COLOR_RANGE_UNKNOWN", "unknown" },
    { C_ENUM(GST_VIDEO_COLOR_RANGE_0_255), "GST_VIDEO_COLOR_RANGE_0_255", "0-255" },
    { C_ENUM(GST_VIDEO_COLOR_RANGE_16_235), "GST_VIDEO_COLOR_RANGE_16_235", "16-235" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoColorRange"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_color_matrix_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_COLOR_MATRIX_UNKNOWN), "GST_VIDEO_COLOR_MATRIX_UNKNOWN", "unknown" },
    { C_ENUM(GST_VIDEO_COLOR_MATRIX_RGB), "GST_VIDEO_COLOR_MATRIX_RGB", "rgb" },
    { C_ENUM(GST_VIDEO_COLOR_MATRIX_FCC), "GST_VIDEO_COLOR_MATRIX_FCC", "fcc" },
    { C_ENUM(GST_VIDEO_COLOR_MATRIX_BT709), "GST_VIDEO_COLOR_MATRIX_BT709", "bt709" },
    { C_ENUM(GST_VIDEO_COLOR_MATRIX_BT601), "GST_VIDEO_COLOR_MATRIX_BT601", "bt601" },
    { C_ENUM(GST_VIDEO_COLOR_MATRIX_SMPTE240M), "GST_VIDEO_COLOR_MATRIX_SMPTE240M", "smpte240m" },
    { C_ENUM(GST_VIDEO_COLOR_MATRIX_BT2020), "GST_VIDEO_COLOR_MATRIX_BT2020", "bt2020" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoColorMatrix"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_transfer_function_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_TRANSFER_UNKNOWN), "GST_VIDEO_TRANSFER_UNKNOWN", "unknown" },
    { C_ENUM(GST_VIDEO_TRANSFER_GAMMA10), "GST_VIDEO_TRANSFER_GAMMA10", "gamma10" },
    { C_ENUM(GST_VIDEO_TRANSFER_GAMMA18), "GST_VIDEO_TRANSFER_GAMMA18", "gamma18" },
    { C_ENUM(GST_VIDEO_TRANSFER_GAMMA20), "GST_VIDEO_TRANSFER_GAMMA20", "gamma20" },
    { C_ENUM(GST_VIDEO_TRANSFER_GAMMA22), "GST_VIDEO_TRANSFER_GAMMA22", "gamma22" },
    { C_ENUM(GST_VIDEO_TRANSFER_BT709), "GST_VIDEO_TRANSFER_BT709", "bt709" },
    { C_ENUM(GST_VIDEO_TRANSFER_SMPTE240M), "GST_VIDEO_TRANSFER_SMPTE240M", "smpte240m" },
    { C_ENUM(GST_VIDEO_TRANSFER_SRGB), "GST_VIDEO_TRANSFER_SRGB", "srgb" },
    { C_ENUM(GST_VIDEO_TRANSFER_GAMMA28), "GST_VIDEO_TRANSFER_GAMMA28", "gamma28" },
    { C_ENUM(GST_VIDEO_TRANSFER_LOG100), "GST_VIDEO_TRANSFER_LOG100", "log100" },
    { C_ENUM(GST_VIDEO_TRANSFER_LOG316), "GST_VIDEO_TRANSFER_LOG316", "log316" },
    { C_ENUM(GST_VIDEO_TRANSFER_BT2020_12), "GST_VIDEO_TRANSFER_BT2020_12", "bt2020-12" },
    { C_ENUM(GST_VIDEO_TRANSFER_ADOBERGB), "GST_VIDEO_TRANSFER_ADOBERGB", "adobergb" },
    { C_ENUM(GST_VIDEO_TRANSFER_BT2020_10), "GST_VIDEO_TRANSFER_BT2020_10", "bt2020-10" },
    { C_ENUM(GST_VIDEO_TRANSFER_SMPTE2084), "GST_VIDEO_TRANSFER_SMPTE2084", "smpte2084" },
    { C_ENUM(GST_VIDEO_TRANSFER_ARIB_STD_B67), "GST_VIDEO_TRANSFER_ARIB_STD_B67", "arib-std-b67" },
    { C_ENUM(GST_VIDEO_TRANSFER_BT601), "GST_VIDEO_TRANSFER_BT601", "bt601" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoTransferFunction"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_color_primaries_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_UNKNOWN), "GST_VIDEO_COLOR_PRIMARIES_UNKNOWN", "unknown" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_BT709), "GST_VIDEO_COLOR_PRIMARIES_BT709", "bt709" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_BT470M), "GST_VIDEO_COLOR_PRIMARIES_BT470M", "bt470m" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_BT470BG), "GST_VIDEO_COLOR_PRIMARIES_BT470BG", "bt470bg" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_SMPTE170M), "GST_VIDEO_COLOR_PRIMARIES_SMPTE170M", "smpte170m" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_SMPTE240M), "GST_VIDEO_COLOR_PRIMARIES_SMPTE240M", "smpte240m" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_FILM), "GST_VIDEO_COLOR_PRIMARIES_FILM", "film" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_BT2020), "GST_VIDEO_COLOR_PRIMARIES_BT2020", "bt2020" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_ADOBERGB), "GST_VIDEO_COLOR_PRIMARIES_ADOBERGB", "adobergb" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_SMPTEST428), "GST_VIDEO_COLOR_PRIMARIES_SMPTEST428", "smptest428" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_SMPTERP431), "GST_VIDEO_COLOR_PRIMARIES_SMPTERP431", "smpterp431" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_SMPTEEG432), "GST_VIDEO_COLOR_PRIMARIES_SMPTEEG432", "smpteeg432" },
    { C_ENUM(GST_VIDEO_COLOR_PRIMARIES_EBU3213), "GST_VIDEO_COLOR_PRIMARIES_EBU3213", "ebu3213" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoColorPrimaries"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-converter.h" */

GType
gst_video_alpha_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_ALPHA_MODE_COPY), "GST_VIDEO_ALPHA_MODE_COPY", "copy" },
    { C_ENUM(GST_VIDEO_ALPHA_MODE_SET), "GST_VIDEO_ALPHA_MODE_SET", "set" },
    { C_ENUM(GST_VIDEO_ALPHA_MODE_MULT), "GST_VIDEO_ALPHA_MODE_MULT", "mult" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoAlphaMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_chroma_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_CHROMA_MODE_FULL), "GST_VIDEO_CHROMA_MODE_FULL", "full" },
    { C_ENUM(GST_VIDEO_CHROMA_MODE_UPSAMPLE_ONLY), "GST_VIDEO_CHROMA_MODE_UPSAMPLE_ONLY", "upsample-only" },
    { C_ENUM(GST_VIDEO_CHROMA_MODE_DOWNSAMPLE_ONLY), "GST_VIDEO_CHROMA_MODE_DOWNSAMPLE_ONLY", "downsample-only" },
    { C_ENUM(GST_VIDEO_CHROMA_MODE_NONE), "GST_VIDEO_CHROMA_MODE_NONE", "none" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoChromaMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_matrix_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_MATRIX_MODE_FULL), "GST_VIDEO_MATRIX_MODE_FULL", "full" },
    { C_ENUM(GST_VIDEO_MATRIX_MODE_INPUT_ONLY), "GST_VIDEO_MATRIX_MODE_INPUT_ONLY", "input-only" },
    { C_ENUM(GST_VIDEO_MATRIX_MODE_OUTPUT_ONLY), "GST_VIDEO_MATRIX_MODE_OUTPUT_ONLY", "output-only" },
    { C_ENUM(GST_VIDEO_MATRIX_MODE_NONE), "GST_VIDEO_MATRIX_MODE_NONE", "none" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoMatrixMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_gamma_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_GAMMA_MODE_NONE), "GST_VIDEO_GAMMA_MODE_NONE", "none" },
    { C_ENUM(GST_VIDEO_GAMMA_MODE_REMAP), "GST_VIDEO_GAMMA_MODE_REMAP", "remap" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoGammaMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_primaries_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_PRIMARIES_MODE_NONE), "GST_VIDEO_PRIMARIES_MODE_NONE", "none" },
    { C_ENUM(GST_VIDEO_PRIMARIES_MODE_MERGE_ONLY), "GST_VIDEO_PRIMARIES_MODE_MERGE_ONLY", "merge-only" },
    { C_ENUM(GST_VIDEO_PRIMARIES_MODE_FAST), "GST_VIDEO_PRIMARIES_MODE_FAST", "fast" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoPrimariesMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-dither.h" */

GType
gst_video_dither_method_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_DITHER_NONE), "GST_VIDEO_DITHER_NONE", "none" },
    { C_ENUM(GST_VIDEO_DITHER_VERTERR), "GST_VIDEO_DITHER_VERTERR", "verterr" },
    { C_ENUM(GST_VIDEO_DITHER_FLOYD_STEINBERG), "GST_VIDEO_DITHER_FLOYD_STEINBERG", "floyd-steinberg" },
    { C_ENUM(GST_VIDEO_DITHER_SIERRA_LITE), "GST_VIDEO_DITHER_SIERRA_LITE", "sierra-lite" },
    { C_ENUM(GST_VIDEO_DITHER_BAYER), "GST_VIDEO_DITHER_BAYER", "bayer" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoDitherMethod"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_dither_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_DITHER_FLAG_NONE), "GST_VIDEO_DITHER_FLAG_NONE", "none" },
    { C_FLAGS(GST_VIDEO_DITHER_FLAG_INTERLACED), "GST_VIDEO_DITHER_FLAG_INTERLACED", "interlaced" },
    { C_FLAGS(GST_VIDEO_DITHER_FLAG_QUANTIZE), "GST_VIDEO_DITHER_FLAG_QUANTIZE", "quantize" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoDitherFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-format.h" */

GType
gst_video_format_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_FORMAT_UNKNOWN), "GST_VIDEO_FORMAT_UNKNOWN", "unknown" },
    { C_ENUM(GST_VIDEO_FORMAT_ENCODED), "GST_VIDEO_FORMAT_ENCODED", "encoded" },
    { C_ENUM(GST_VIDEO_FORMAT_I420), "GST_VIDEO_FORMAT_I420", "i420" },
    { C_ENUM(GST_VIDEO_FORMAT_YV12), "GST_VIDEO_FORMAT_YV12", "yv12" },
    { C_ENUM(GST_VIDEO_FORMAT_YUY2), "GST_VIDEO_FORMAT_YUY2", "yuy2" },
    { C_ENUM(GST_VIDEO_FORMAT_UYVY), "GST_VIDEO_FORMAT_UYVY", "uyvy" },
    { C_ENUM(GST_VIDEO_FORMAT_AYUV), "GST_VIDEO_FORMAT_AYUV", "ayuv" },
    { C_ENUM(GST_VIDEO_FORMAT_RGBx), "GST_VIDEO_FORMAT_RGBx", "rgbx" },
    { C_ENUM(GST_VIDEO_FORMAT_BGRx), "GST_VIDEO_FORMAT_BGRx", "bgrx" },
    { C_ENUM(GST_VIDEO_FORMAT_xRGB), "GST_VIDEO_FORMAT_xRGB", "xrgb" },
    { C_ENUM(GST_VIDEO_FORMAT_xBGR), "GST_VIDEO_FORMAT_xBGR", "xbgr" },
    { C_ENUM(GST_VIDEO_FORMAT_RGBA), "GST_VIDEO_FORMAT_RGBA", "rgba" },
    { C_ENUM(GST_VIDEO_FORMAT_BGRA), "GST_VIDEO_FORMAT_BGRA", "bgra" },
    { C_ENUM(GST_VIDEO_FORMAT_ARGB), "GST_VIDEO_FORMAT_ARGB", "argb" },
    { C_ENUM(GST_VIDEO_FORMAT_ABGR), "GST_VIDEO_FORMAT_ABGR", "abgr" },
    { C_ENUM(GST_VIDEO_FORMAT_RGB), "GST_VIDEO_FORMAT_RGB", "rgb" },
    { C_ENUM(GST_VIDEO_FORMAT_BGR), "GST_VIDEO_FORMAT_BGR", "bgr" },
    { C_ENUM(GST_VIDEO_FORMAT_Y41B), "GST_VIDEO_FORMAT_Y41B", "y41b" },
    { C_ENUM(GST_VIDEO_FORMAT_Y42B), "GST_VIDEO_FORMAT_Y42B", "y42b" },
    { C_ENUM(GST_VIDEO_FORMAT_YVYU), "GST_VIDEO_FORMAT_YVYU", "yvyu" },
    { C_ENUM(GST_VIDEO_FORMAT_Y444), "GST_VIDEO_FORMAT_Y444", "y444" },
    { C_ENUM(GST_VIDEO_FORMAT_v210), "GST_VIDEO_FORMAT_v210", "v210" },
    { C_ENUM(GST_VIDEO_FORMAT_v216), "GST_VIDEO_FORMAT_v216", "v216" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12), "GST_VIDEO_FORMAT_NV12", "nv12" },
    { C_ENUM(GST_VIDEO_FORMAT_NV21), "GST_VIDEO_FORMAT_NV21", "nv21" },
    { C_ENUM(GST_VIDEO_FORMAT_GRAY8), "GST_VIDEO_FORMAT_GRAY8", "gray8" },
    { C_ENUM(GST_VIDEO_FORMAT_GRAY16_BE), "GST_VIDEO_FORMAT_GRAY16_BE", "gray16-be" },
    { C_ENUM(GST_VIDEO_FORMAT_GRAY16_LE), "GST_VIDEO_FORMAT_GRAY16_LE", "gray16-le" },
    { C_ENUM(GST_VIDEO_FORMAT_v308), "GST_VIDEO_FORMAT_v308", "v308" },
    { C_ENUM(GST_VIDEO_FORMAT_RGB16), "GST_VIDEO_FORMAT_RGB16", "rgb16" },
    { C_ENUM(GST_VIDEO_FORMAT_BGR16), "GST_VIDEO_FORMAT_BGR16", "bgr16" },
    { C_ENUM(GST_VIDEO_FORMAT_RGB15), "GST_VIDEO_FORMAT_RGB15", "rgb15" },
    { C_ENUM(GST_VIDEO_FORMAT_BGR15), "GST_VIDEO_FORMAT_BGR15", "bgr15" },
    { C_ENUM(GST_VIDEO_FORMAT_UYVP), "GST_VIDEO_FORMAT_UYVP", "uyvp" },
    { C_ENUM(GST_VIDEO_FORMAT_A420), "GST_VIDEO_FORMAT_A420", "a420" },
    { C_ENUM(GST_VIDEO_FORMAT_RGB8P), "GST_VIDEO_FORMAT_RGB8P", "rgb8p" },
    { C_ENUM(GST_VIDEO_FORMAT_YUV9), "GST_VIDEO_FORMAT_YUV9", "yuv9" },
    { C_ENUM(GST_VIDEO_FORMAT_YVU9), "GST_VIDEO_FORMAT_YVU9", "yvu9" },
    { C_ENUM(GST_VIDEO_FORMAT_IYU1), "GST_VIDEO_FORMAT_IYU1", "iyu1" },
    { C_ENUM(GST_VIDEO_FORMAT_ARGB64), "GST_VIDEO_FORMAT_ARGB64", "argb64" },
    { C_ENUM(GST_VIDEO_FORMAT_AYUV64), "GST_VIDEO_FORMAT_AYUV64", "ayuv64" },
    { C_ENUM(GST_VIDEO_FORMAT_r210), "GST_VIDEO_FORMAT_r210", "r210" },
    { C_ENUM(GST_VIDEO_FORMAT_I420_10BE), "GST_VIDEO_FORMAT_I420_10BE", "i420-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_I420_10LE), "GST_VIDEO_FORMAT_I420_10LE", "i420-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_I422_10BE), "GST_VIDEO_FORMAT_I422_10BE", "i422-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_I422_10LE), "GST_VIDEO_FORMAT_I422_10LE", "i422-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_Y444_10BE), "GST_VIDEO_FORMAT_Y444_10BE", "y444-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_Y444_10LE), "GST_VIDEO_FORMAT_Y444_10LE", "y444-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_GBR), "GST_VIDEO_FORMAT_GBR", "gbr" },
    { C_ENUM(GST_VIDEO_FORMAT_GBR_10BE), "GST_VIDEO_FORMAT_GBR_10BE", "gbr-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_GBR_10LE), "GST_VIDEO_FORMAT_GBR_10LE", "gbr-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_NV16), "GST_VIDEO_FORMAT_NV16", "nv16" },
    { C_ENUM(GST_VIDEO_FORMAT_NV24), "GST_VIDEO_FORMAT_NV24", "nv24" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12_64Z32), "GST_VIDEO_FORMAT_NV12_64Z32", "nv12-64z32" },
    { C_ENUM(GST_VIDEO_FORMAT_A420_10BE), "GST_VIDEO_FORMAT_A420_10BE", "a420-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_A420_10LE), "GST_VIDEO_FORMAT_A420_10LE", "a420-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_A422_10BE), "GST_VIDEO_FORMAT_A422_10BE", "a422-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_A422_10LE), "GST_VIDEO_FORMAT_A422_10LE", "a422-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_A444_10BE), "GST_VIDEO_FORMAT_A444_10BE", "a444-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_A444_10LE), "GST_VIDEO_FORMAT_A444_10LE", "a444-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_NV61), "GST_VIDEO_FORMAT_NV61", "nv61" },
    { C_ENUM(GST_VIDEO_FORMAT_P010_10BE), "GST_VIDEO_FORMAT_P010_10BE", "p010-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_P010_10LE), "GST_VIDEO_FORMAT_P010_10LE", "p010-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_P210_10LE), "GST_VIDEO_FORMAT_P210_10LE", "p210-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_P212_12LE), "GST_VIDEO_FORMAT_P212_12LE", "p212-12le" },
    { C_ENUM(GST_VIDEO_FORMAT_IYU2), "GST_VIDEO_FORMAT_IYU2", "iyu2" },
    { C_ENUM(GST_VIDEO_FORMAT_VYUY), "GST_VIDEO_FORMAT_VYUY", "vyuy" },
    { C_ENUM(GST_VIDEO_FORMAT_GBRA), "GST_VIDEO_FORMAT_GBRA", "gbra" },
    { C_ENUM(GST_VIDEO_FORMAT_GBRA_10BE), "GST_VIDEO_FORMAT_GBRA_10BE", "gbra-10be" },
    { C_ENUM(GST_VIDEO_FORMAT_GBRA_10LE), "GST_VIDEO_FORMAT_GBRA_10LE", "gbra-10le" },
    { C_ENUM(GST_VIDEO_FORMAT_GBR_12BE), "GST_VIDEO_FORMAT_GBR_12BE", "gbr-12be" },
    { C_ENUM(GST_VIDEO_FORMAT_GBR_12LE), "GST_VIDEO_FORMAT_GBR_12LE", "gbr-12le" },
    { C_ENUM(GST_VIDEO_FORMAT_GBRA_12BE), "GST_VIDEO_FORMAT_GBRA_12BE", "gbra-12be" },
    { C_ENUM(GST_VIDEO_FORMAT_GBRA_12LE), "GST_VIDEO_FORMAT_GBRA_12LE", "gbra-12le" },
    { C_ENUM(GST_VIDEO_FORMAT_I420_12BE), "GST_VIDEO_FORMAT_I420_12BE", "i420-12be" },
    { C_ENUM(GST_VIDEO_FORMAT_I420_12LE), "GST_VIDEO_FORMAT_I420_12LE", "i420-12le" },
    { C_ENUM(GST_VIDEO_FORMAT_I422_12BE), "GST_VIDEO_FORMAT_I422_12BE", "i422-12be" },
    { C_ENUM(GST_VIDEO_FORMAT_I422_12LE), "GST_VIDEO_FORMAT_I422_12LE", "i422-12le" },
    { C_ENUM(GST_VIDEO_FORMAT_Y444_12BE), "GST_VIDEO_FORMAT_Y444_12BE", "y444-12be" },
    { C_ENUM(GST_VIDEO_FORMAT_Y444_12LE), "GST_VIDEO_FORMAT_Y444_12LE", "y444-12le" },
    { C_ENUM(GST_VIDEO_FORMAT_GRAY10_LE32), "GST_VIDEO_FORMAT_GRAY10_LE32", "gray10-le32" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12_10LE32), "GST_VIDEO_FORMAT_NV12_10LE32", "nv12-10le32" },
    { C_ENUM(GST_VIDEO_FORMAT_NV16_10LE32), "GST_VIDEO_FORMAT_NV16_10LE32", "nv16-10le32" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12_10LE40), "GST_VIDEO_FORMAT_NV12_10LE40", "nv12-10le40" },
    { C_ENUM(GST_VIDEO_FORMAT_Y210), "GST_VIDEO_FORMAT_Y210", "y210" },
    { C_ENUM(GST_VIDEO_FORMAT_Y410), "GST_VIDEO_FORMAT_Y410", "y410" },
    { C_ENUM(GST_VIDEO_FORMAT_VUYA), "GST_VIDEO_FORMAT_VUYA", "vuya" },
    { C_ENUM(GST_VIDEO_FORMAT_BGR10A2_LE), "GST_VIDEO_FORMAT_BGR10A2_LE", "bgr10a2-le" },
    { C_ENUM(GST_VIDEO_FORMAT_RGB10A2_LE), "GST_VIDEO_FORMAT_RGB10A2_LE", "rgb10a2-le" },
    { C_ENUM(GST_VIDEO_FORMAT_Y444_16BE), "GST_VIDEO_FORMAT_Y444_16BE", "y444-16be" },
    { C_ENUM(GST_VIDEO_FORMAT_Y444_16LE), "GST_VIDEO_FORMAT_Y444_16LE", "y444-16le" },
    { C_ENUM(GST_VIDEO_FORMAT_P016_BE), "GST_VIDEO_FORMAT_P016_BE", "p016-be" },
    { C_ENUM(GST_VIDEO_FORMAT_P016_LE), "GST_VIDEO_FORMAT_P016_LE", "p016-le" },
    { C_ENUM(GST_VIDEO_FORMAT_P012_BE), "GST_VIDEO_FORMAT_P012_BE", "p012-be" },
    { C_ENUM(GST_VIDEO_FORMAT_P012_LE), "GST_VIDEO_FORMAT_P012_LE", "p012-le" },
    { C_ENUM(GST_VIDEO_FORMAT_Y212_BE), "GST_VIDEO_FORMAT_Y212_BE", "y212-be" },
    { C_ENUM(GST_VIDEO_FORMAT_Y212_LE), "GST_VIDEO_FORMAT_Y212_LE", "y212-le" },
    { C_ENUM(GST_VIDEO_FORMAT_Y412_BE), "GST_VIDEO_FORMAT_Y412_BE", "y412-be" },
    { C_ENUM(GST_VIDEO_FORMAT_Y412_LE), "GST_VIDEO_FORMAT_Y412_LE", "y412-le" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12_4L4), "GST_VIDEO_FORMAT_NV12_4L4", "nv12-4l4" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12_32L32), "GST_VIDEO_FORMAT_NV12_32L32", "nv12-32l32" },
    { C_ENUM(GST_VIDEO_FORMAT_Y444_10LE32), "GST_VIDEO_FORMAT_Y444_10LE32", "y444-10le32" },
    { C_ENUM(GST_VIDEO_FORMAT_RGBP), "GST_VIDEO_FORMAT_RGBP", "rgbp" },
    { C_ENUM(GST_VIDEO_FORMAT_BGRP), "GST_VIDEO_FORMAT_BGRP", "bgrp" },
    { C_ENUM(GST_VIDEO_FORMAT_AV12), "GST_VIDEO_FORMAT_AV12", "av12" },
    { C_ENUM(GST_VIDEO_FORMAT_ARGB64_LE), "GST_VIDEO_FORMAT_ARGB64_LE", "argb64-le" },
    { C_ENUM(GST_VIDEO_FORMAT_ARGB64_BE), "GST_VIDEO_FORMAT_ARGB64_BE", "argb64-be" },
    { C_ENUM(GST_VIDEO_FORMAT_RGBA64_LE), "GST_VIDEO_FORMAT_RGBA64_LE", "rgba64-le" },
    { C_ENUM(GST_VIDEO_FORMAT_RGBA64_BE), "GST_VIDEO_FORMAT_RGBA64_BE", "rgba64-be" },
    { C_ENUM(GST_VIDEO_FORMAT_BGRA64_LE), "GST_VIDEO_FORMAT_BGRA64_LE", "bgra64-le" },
    { C_ENUM(GST_VIDEO_FORMAT_BGRA64_BE), "GST_VIDEO_FORMAT_BGRA64_BE", "bgra64-be" },
    { C_ENUM(GST_VIDEO_FORMAT_ABGR64_LE), "GST_VIDEO_FORMAT_ABGR64_LE", "abgr64-le" },
    { C_ENUM(GST_VIDEO_FORMAT_ABGR64_BE), "GST_VIDEO_FORMAT_ABGR64_BE", "abgr64-be" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12_16L32S), "GST_VIDEO_FORMAT_NV12_16L32S", "nv12-16l32s" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12_8L128), "GST_VIDEO_FORMAT_NV12_8L128", "nv12-8l128" },
    { C_ENUM(GST_VIDEO_FORMAT_NV12_10BE_8L128), "GST_VIDEO_FORMAT_NV12_10BE_8L128", "nv12-10be-8l128" },
    { C_ENUM(GST_VIDEO_FORMAT_GRAY10_LE), "GST_VIDEO_FORMAT_GRAY10_LE", "gray10-le" },
    { C_ENUM(GST_VIDEO_FORMAT_GRAY12_LE), "GST_VIDEO_FORMAT_GRAY12_LE", "gray12-le" },
    { C_ENUM(GST_VIDEO_FORMAT_T5M8), "GST_VIDEO_FORMAT_T5M8", "t5m8" },
    { C_ENUM(GST_VIDEO_FORMAT_T5MA), "GST_VIDEO_FORMAT_T5MA", "t5ma" },
    { C_ENUM(GST_VIDEO_FORMAT_T5MC), "GST_VIDEO_FORMAT_T5MC", "t5mc" },
    { C_ENUM(GST_VIDEO_FORMAT_T508), "GST_VIDEO_FORMAT_T508", "t508" },
    { C_ENUM(GST_VIDEO_FORMAT_T50A), "GST_VIDEO_FORMAT_T50A", "t50a" },
    { C_ENUM(GST_VIDEO_FORMAT_T50C), "GST_VIDEO_FORMAT_T50C", "t50c" },
    { C_ENUM(GST_VIDEO_FORMAT_T528), "GST_VIDEO_FORMAT_T528", "t528" },
    { C_ENUM(GST_VIDEO_FORMAT_T52A), "GST_VIDEO_FORMAT_T52A", "t52a" },
    { C_ENUM(GST_VIDEO_FORMAT_T52C), "GST_VIDEO_FORMAT_T52C", "t52c" },
    { C_ENUM(GST_VIDEO_FORMAT_T548), "GST_VIDEO_FORMAT_T548", "t548" },
    { C_ENUM(GST_VIDEO_FORMAT_T54A), "GST_VIDEO_FORMAT_T54A", "t54a" },
    { C_ENUM(GST_VIDEO_FORMAT_T54C), "GST_VIDEO_FORMAT_T54C", "t54c" },
    { C_ENUM(GST_VIDEO_FORMAT_T6M8), "GST_VIDEO_FORMAT_T6M8", "t6m8" },
    { C_ENUM(GST_VIDEO_FORMAT_T6MA), "GST_VIDEO_FORMAT_T6MA", "t6ma" },
    { C_ENUM(GST_VIDEO_FORMAT_T6MC), "GST_VIDEO_FORMAT_T6MC", "t6mc" },
    { C_ENUM(GST_VIDEO_FORMAT_T608), "GST_VIDEO_FORMAT_T608", "t608" },
    { C_ENUM(GST_VIDEO_FORMAT_T60A), "GST_VIDEO_FORMAT_T60A", "t60a" },
    { C_ENUM(GST_VIDEO_FORMAT_T60C), "GST_VIDEO_FORMAT_T60C", "t60c" },
    { C_ENUM(GST_VIDEO_FORMAT_T628), "GST_VIDEO_FORMAT_T628", "t628" },
    { C_ENUM(GST_VIDEO_FORMAT_T62A), "GST_VIDEO_FORMAT_T62A", "t62a" },
    { C_ENUM(GST_VIDEO_FORMAT_T62C), "GST_VIDEO_FORMAT_T62C", "t62c" },
    { C_ENUM(GST_VIDEO_FORMAT_T648), "GST_VIDEO_FORMAT_T648", "t648" },
    { C_ENUM(GST_VIDEO_FORMAT_T64A), "GST_VIDEO_FORMAT_T64A", "t64a" },
    { C_ENUM(GST_VIDEO_FORMAT_T64C), "GST_VIDEO_FORMAT_T64C", "t64c" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoFormat"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_format_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_YUV), "GST_VIDEO_FORMAT_FLAG_YUV", "yuv" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_RGB), "GST_VIDEO_FORMAT_FLAG_RGB", "rgb" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_GRAY), "GST_VIDEO_FORMAT_FLAG_GRAY", "gray" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_ALPHA), "GST_VIDEO_FORMAT_FLAG_ALPHA", "alpha" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_LE), "GST_VIDEO_FORMAT_FLAG_LE", "le" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_PALETTE), "GST_VIDEO_FORMAT_FLAG_PALETTE", "palette" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_COMPLEX), "GST_VIDEO_FORMAT_FLAG_COMPLEX", "complex" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_UNPACK), "GST_VIDEO_FORMAT_FLAG_UNPACK", "unpack" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_TILED), "GST_VIDEO_FORMAT_FLAG_TILED", "tiled" },
    { C_FLAGS(GST_VIDEO_FORMAT_FLAG_SUBTILES), "GST_VIDEO_FORMAT_FLAG_SUBTILES", "subtiles" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoFormatFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_pack_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_PACK_FLAG_NONE), "GST_VIDEO_PACK_FLAG_NONE", "none" },
    { C_FLAGS(GST_VIDEO_PACK_FLAG_TRUNCATE_RANGE), "GST_VIDEO_PACK_FLAG_TRUNCATE_RANGE", "truncate-range" },
    { C_FLAGS(GST_VIDEO_PACK_FLAG_INTERLACED), "GST_VIDEO_PACK_FLAG_INTERLACED", "interlaced" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoPackFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-frame.h" */

GType
gst_video_frame_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_NONE), "GST_VIDEO_FRAME_FLAG_NONE", "none" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_INTERLACED), "GST_VIDEO_FRAME_FLAG_INTERLACED", "interlaced" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_TFF), "GST_VIDEO_FRAME_FLAG_TFF", "tff" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_RFF), "GST_VIDEO_FRAME_FLAG_RFF", "rff" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_ONEFIELD), "GST_VIDEO_FRAME_FLAG_ONEFIELD", "onefield" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_MULTIPLE_VIEW), "GST_VIDEO_FRAME_FLAG_MULTIPLE_VIEW", "multiple-view" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_FIRST_IN_BUNDLE), "GST_VIDEO_FRAME_FLAG_FIRST_IN_BUNDLE", "first-in-bundle" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_TOP_FIELD), "GST_VIDEO_FRAME_FLAG_TOP_FIELD", "top-field" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_ONEFIELD), "GST_VIDEO_FRAME_FLAG_ONEFIELD", "onefield" },
    { C_FLAGS(GST_VIDEO_FRAME_FLAG_BOTTOM_FIELD), "GST_VIDEO_FRAME_FLAG_BOTTOM_FIELD", "bottom-field" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoFrameFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_buffer_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_INTERLACED), "GST_VIDEO_BUFFER_FLAG_INTERLACED", "interlaced" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_TFF), "GST_VIDEO_BUFFER_FLAG_TFF", "tff" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_RFF), "GST_VIDEO_BUFFER_FLAG_RFF", "rff" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_ONEFIELD), "GST_VIDEO_BUFFER_FLAG_ONEFIELD", "onefield" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_MULTIPLE_VIEW), "GST_VIDEO_BUFFER_FLAG_MULTIPLE_VIEW", "multiple-view" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_FIRST_IN_BUNDLE), "GST_VIDEO_BUFFER_FLAG_FIRST_IN_BUNDLE", "first-in-bundle" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_TOP_FIELD), "GST_VIDEO_BUFFER_FLAG_TOP_FIELD", "top-field" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_ONEFIELD), "GST_VIDEO_BUFFER_FLAG_ONEFIELD", "onefield" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_BOTTOM_FIELD), "GST_VIDEO_BUFFER_FLAG_BOTTOM_FIELD", "bottom-field" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_MARKER), "GST_VIDEO_BUFFER_FLAG_MARKER", "marker" },
    { C_FLAGS(GST_VIDEO_BUFFER_FLAG_LAST), "GST_VIDEO_BUFFER_FLAG_LAST", "last" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoBufferFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_frame_map_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_FRAME_MAP_FLAG_NO_REF), "GST_VIDEO_FRAME_MAP_FLAG_NO_REF", "no-ref" },
    { C_FLAGS(GST_VIDEO_FRAME_MAP_FLAG_LAST), "GST_VIDEO_FRAME_MAP_FLAG_LAST", "last" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoFrameMapFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-info.h" */

GType
gst_video_interlace_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_INTERLACE_MODE_PROGRESSIVE), "GST_VIDEO_INTERLACE_MODE_PROGRESSIVE", "progressive" },
    { C_ENUM(GST_VIDEO_INTERLACE_MODE_INTERLEAVED), "GST_VIDEO_INTERLACE_MODE_INTERLEAVED", "interleaved" },
    { C_ENUM(GST_VIDEO_INTERLACE_MODE_MIXED), "GST_VIDEO_INTERLACE_MODE_MIXED", "mixed" },
    { C_ENUM(GST_VIDEO_INTERLACE_MODE_FIELDS), "GST_VIDEO_INTERLACE_MODE_FIELDS", "fields" },
    { C_ENUM(GST_VIDEO_INTERLACE_MODE_ALTERNATE), "GST_VIDEO_INTERLACE_MODE_ALTERNATE", "alternate" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoInterlaceMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_multiview_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_NONE), "GST_VIDEO_MULTIVIEW_MODE_NONE", "none" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_MONO), "GST_VIDEO_MULTIVIEW_MODE_MONO", "mono" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_LEFT), "GST_VIDEO_MULTIVIEW_MODE_LEFT", "left" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_RIGHT), "GST_VIDEO_MULTIVIEW_MODE_RIGHT", "right" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_SIDE_BY_SIDE), "GST_VIDEO_MULTIVIEW_MODE_SIDE_BY_SIDE", "side-by-side" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_SIDE_BY_SIDE_QUINCUNX), "GST_VIDEO_MULTIVIEW_MODE_SIDE_BY_SIDE_QUINCUNX", "side-by-side-quincunx" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_COLUMN_INTERLEAVED), "GST_VIDEO_MULTIVIEW_MODE_COLUMN_INTERLEAVED", "column-interleaved" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_ROW_INTERLEAVED), "GST_VIDEO_MULTIVIEW_MODE_ROW_INTERLEAVED", "row-interleaved" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_TOP_BOTTOM), "GST_VIDEO_MULTIVIEW_MODE_TOP_BOTTOM", "top-bottom" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_CHECKERBOARD), "GST_VIDEO_MULTIVIEW_MODE_CHECKERBOARD", "checkerboard" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_FRAME_BY_FRAME), "GST_VIDEO_MULTIVIEW_MODE_FRAME_BY_FRAME", "frame-by-frame" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_MULTIVIEW_FRAME_BY_FRAME), "GST_VIDEO_MULTIVIEW_MODE_MULTIVIEW_FRAME_BY_FRAME", "multiview-frame-by-frame" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_MODE_SEPARATED), "GST_VIDEO_MULTIVIEW_MODE_SEPARATED", "separated" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoMultiviewMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_multiview_frame_packing_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_NONE), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_NONE", "none" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_MONO), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_MONO", "mono" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_LEFT), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_LEFT", "left" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_RIGHT), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_RIGHT", "right" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_SIDE_BY_SIDE), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_SIDE_BY_SIDE", "side-by-side" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_SIDE_BY_SIDE_QUINCUNX), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_SIDE_BY_SIDE_QUINCUNX", "side-by-side-quincunx" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_COLUMN_INTERLEAVED), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_COLUMN_INTERLEAVED", "column-interleaved" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_ROW_INTERLEAVED), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_ROW_INTERLEAVED", "row-interleaved" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_TOP_BOTTOM), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_TOP_BOTTOM", "top-bottom" },
    { C_ENUM(GST_VIDEO_MULTIVIEW_FRAME_PACKING_CHECKERBOARD), "GST_VIDEO_MULTIVIEW_FRAME_PACKING_CHECKERBOARD", "checkerboard" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoMultiviewFramePacking"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_multiview_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_MULTIVIEW_FLAGS_NONE), "GST_VIDEO_MULTIVIEW_FLAGS_NONE", "none" },
    { C_FLAGS(GST_VIDEO_MULTIVIEW_FLAGS_RIGHT_VIEW_FIRST), "GST_VIDEO_MULTIVIEW_FLAGS_RIGHT_VIEW_FIRST", "right-view-first" },
    { C_FLAGS(GST_VIDEO_MULTIVIEW_FLAGS_LEFT_FLIPPED), "GST_VIDEO_MULTIVIEW_FLAGS_LEFT_FLIPPED", "left-flipped" },
    { C_FLAGS(GST_VIDEO_MULTIVIEW_FLAGS_LEFT_FLOPPED), "GST_VIDEO_MULTIVIEW_FLAGS_LEFT_FLOPPED", "left-flopped" },
    { C_FLAGS(GST_VIDEO_MULTIVIEW_FLAGS_RIGHT_FLIPPED), "GST_VIDEO_MULTIVIEW_FLAGS_RIGHT_FLIPPED", "right-flipped" },
    { C_FLAGS(GST_VIDEO_MULTIVIEW_FLAGS_RIGHT_FLOPPED), "GST_VIDEO_MULTIVIEW_FLAGS_RIGHT_FLOPPED", "right-flopped" },
    { C_FLAGS(GST_VIDEO_MULTIVIEW_FLAGS_HALF_ASPECT), "GST_VIDEO_MULTIVIEW_FLAGS_HALF_ASPECT", "half-aspect" },
    { C_FLAGS(GST_VIDEO_MULTIVIEW_FLAGS_MIXED_MONO), "GST_VIDEO_MULTIVIEW_FLAGS_MIXED_MONO", "mixed-mono" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoMultiviewFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_FLAG_NONE), "GST_VIDEO_FLAG_NONE", "none" },
    { C_FLAGS(GST_VIDEO_FLAG_VARIABLE_FPS), "GST_VIDEO_FLAG_VARIABLE_FPS", "variable-fps" },
    { C_FLAGS(GST_VIDEO_FLAG_PREMULTIPLIED_ALPHA), "GST_VIDEO_FLAG_PREMULTIPLIED_ALPHA", "premultiplied-alpha" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_field_order_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_FIELD_ORDER_UNKNOWN), "GST_VIDEO_FIELD_ORDER_UNKNOWN", "unknown" },
    { C_ENUM(GST_VIDEO_FIELD_ORDER_TOP_FIELD_FIRST), "GST_VIDEO_FIELD_ORDER_TOP_FIELD_FIRST", "top-field-first" },
    { C_ENUM(GST_VIDEO_FIELD_ORDER_BOTTOM_FIELD_FIRST), "GST_VIDEO_FIELD_ORDER_BOTTOM_FIELD_FIRST", "bottom-field-first" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoFieldOrder"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-overlay-composition.h" */

GType
gst_video_overlay_format_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_OVERLAY_FORMAT_FLAG_NONE), "GST_VIDEO_OVERLAY_FORMAT_FLAG_NONE", "none" },
    { C_FLAGS(GST_VIDEO_OVERLAY_FORMAT_FLAG_PREMULTIPLIED_ALPHA), "GST_VIDEO_OVERLAY_FORMAT_FLAG_PREMULTIPLIED_ALPHA", "premultiplied-alpha" },
    { C_FLAGS(GST_VIDEO_OVERLAY_FORMAT_FLAG_GLOBAL_ALPHA), "GST_VIDEO_OVERLAY_FORMAT_FLAG_GLOBAL_ALPHA", "global-alpha" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoOverlayFormatFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-resampler.h" */

GType
gst_video_resampler_method_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_RESAMPLER_METHOD_NEAREST), "GST_VIDEO_RESAMPLER_METHOD_NEAREST", "nearest" },
    { C_ENUM(GST_VIDEO_RESAMPLER_METHOD_LINEAR), "GST_VIDEO_RESAMPLER_METHOD_LINEAR", "linear" },
    { C_ENUM(GST_VIDEO_RESAMPLER_METHOD_CUBIC), "GST_VIDEO_RESAMPLER_METHOD_CUBIC", "cubic" },
    { C_ENUM(GST_VIDEO_RESAMPLER_METHOD_SINC), "GST_VIDEO_RESAMPLER_METHOD_SINC", "sinc" },
    { C_ENUM(GST_VIDEO_RESAMPLER_METHOD_LANCZOS), "GST_VIDEO_RESAMPLER_METHOD_LANCZOS", "lanczos" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoResamplerMethod"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_resampler_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_RESAMPLER_FLAG_NONE), "GST_VIDEO_RESAMPLER_FLAG_NONE", "none" },
    { C_FLAGS(GST_VIDEO_RESAMPLER_FLAG_HALF_TAPS), "GST_VIDEO_RESAMPLER_FLAG_HALF_TAPS", "half-taps" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoResamplerFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-scaler.h" */

GType
gst_video_scaler_flags_get_type (void)
{
static gsize gtype_id = 0;
static const GFlagsValue values[] = {
    { C_FLAGS(GST_VIDEO_SCALER_FLAG_NONE), "GST_VIDEO_SCALER_FLAG_NONE", "none" },
    { C_FLAGS(GST_VIDEO_SCALER_FLAG_INTERLACED), "GST_VIDEO_SCALER_FLAG_INTERLACED", "interlaced" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_flags_register_static (g_intern_static_string ("GstVideoScalerFlags"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video-tile.h" */

GType
gst_video_tile_type_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_TILE_TYPE_INDEXED), "GST_VIDEO_TILE_TYPE_INDEXED", "indexed" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoTileType"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
gst_video_tile_mode_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_TILE_MODE_UNKNOWN), "GST_VIDEO_TILE_MODE_UNKNOWN", "unknown" },
    { C_ENUM(GST_VIDEO_TILE_MODE_ZFLIPZ_2X2), "GST_VIDEO_TILE_MODE_ZFLIPZ_2X2", "zflipz-2x2" },
    { C_ENUM(GST_VIDEO_TILE_MODE_LINEAR), "GST_VIDEO_TILE_MODE_LINEAR", "linear" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoTileMode"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "video.h" */

GType
gst_video_orientation_method_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(GST_VIDEO_ORIENTATION_IDENTITY), "GST_VIDEO_ORIENTATION_IDENTITY", "identity" },
    { C_ENUM(GST_VIDEO_ORIENTATION_90R), "GST_VIDEO_ORIENTATION_90R", "90r" },
    { C_ENUM(GST_VIDEO_ORIENTATION_180), "GST_VIDEO_ORIENTATION_180", "180" },
    { C_ENUM(GST_VIDEO_ORIENTATION_90L), "GST_VIDEO_ORIENTATION_90L", "90l" },
    { C_ENUM(GST_VIDEO_ORIENTATION_HORIZ), "GST_VIDEO_ORIENTATION_HORIZ", "horiz" },
    { C_ENUM(GST_VIDEO_ORIENTATION_VERT), "GST_VIDEO_ORIENTATION_VERT", "vert" },
    { C_ENUM(GST_VIDEO_ORIENTATION_UL_LR), "GST_VIDEO_ORIENTATION_UL_LR", "ul-lr" },
    { C_ENUM(GST_VIDEO_ORIENTATION_UR_LL), "GST_VIDEO_ORIENTATION_UR_LL", "ur-ll" },
    { C_ENUM(GST_VIDEO_ORIENTATION_AUTO), "GST_VIDEO_ORIENTATION_AUTO", "auto" },
    { C_ENUM(GST_VIDEO_ORIENTATION_CUSTOM), "GST_VIDEO_ORIENTATION_CUSTOM", "custom" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("GstVideoOrientationMethod"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* Generated data ends here */

